"""
Routes for live access to CHAPPS state
--------------------------------------

Please note that documentation about API routes is best
viewed in the OpenAPI docs generated by FastAPI_ when
the API is launched.

"""
from typing import List, Optional
from fastapi import status, APIRouter, Body, HTTPException
from sqlalchemy.orm import sessionmaker
from chapps.rest.routers.common import (
    load_model_with_assoc,
    load_models_with_assoc,
)
from chapps.dbsession import sql_engine
from chapps.models import (
    User,
    Domain,
    Email,
    LiveQuotaResp,
    TextResp,
    TimeResp,
    InstanceTimesResp,
    SourceUserMapResp,
    BulkQuotaResp,
    DeleteResp,
    user_quota_assoc,
)
from chapps.policy import (
    OutboundQuotaPolicy,
    SenderDomainAuthPolicy,
    GreylistingPolicy,
)
from chapps.spf_policy import SPFEnforcementPolicy
from chapps.config import CHAPPSConfig
from chapps.util import hash_password
import logging

logger = logging.getLogger(__name__)
config = CHAPPSConfig.get_config()
Session = sessionmaker(sql_engine)

api = APIRouter(
    prefix="/live",
    tags=["live"],
    responses={
        status.HTTP_404_NOT_FOUND: {"description": "Resource not found."},
        status.HTTP_400_BAD_REQUEST: {
            "description": ("The request was missing some required data.")
        },
    },
)
"""The live routes API router"""

# define some useful functions
load_user_with_quota = load_model_with_assoc(User, assoc=[user_quota_assoc])
r"""Create a `load_user_with_quota` function, using the factory.

:param int item_id: ID of user to load, or 0 if loading by name

:param Optiona[str] name: exact name of **User** to load, if `item_id` is zero

:returns: a tuple of (**User**, dict(quota=\ **Quota**), [remarks,...])

Because the factory is meant to work with any number of associations, the
return format is a little clunky for this case but it seems a small price to
pay to avoid code duplication.

Also note that `FastAPI`_ interprets the docstrings of the coroutines defined
in this module, in order to produce the live API documentation.  As a result,
they have been kept simpler than in some other modules.

"""

load_users_with_quota = load_models_with_assoc(User, assoc=user_quota_assoc)
"""Create a username to quota id mapping function

:param item_ids: list of **User** ids to associate to quotas

:returns: a list of user ORM objects with associated quotas

"""


@api.get("/quota/", response_model=BulkQuotaResp)
async def get_bulk_quota_remaining(user_ids: List[int]) -> BulkQuotaResp:
    """Accepts a list of user ids.

    Returns a list of JSON objects with attributes named `user_name` and
    `quota_avail`, representing the remaining number of email transmissions
    available to that user at the moment the query was executed.

    """
    remarks = []
    response = []
    oqp = OutboundQuotaPolicy()
    uqm = load_users_with_quota(user_ids)
    for user in uqm:
        avail, rmks = oqp.current_quota(user.name, user.quota)
        response.append(dict(user_name=user.name, quota_avail=avail))
        remarks.extend(rmks)
    return BulkQuotaResp.send(response, remarks=remarks)


@api.get("/quota/{user_id}", response_model=LiveQuotaResp)
async def get_current_quota_remaining_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
) -> LiveQuotaResp:
    """Accepts the id of the user whose remaining quota should be checked.

    Returns the instantaneous number of available send attempts in
    `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    response, more_remarks = oqp.current_quota(user.name, quota)
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.delete("/quota/{user_id}", response_model=LiveQuotaResp)
async def reset_live_quota_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
) -> LiveQuotaResp:
    """Accepts the id of the user whose quota should be reset.

    Returns the number of send attempts dropped in `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    response, more_remarks = oqp.reset_quota(user.name)
    if not quota:
        remarks.append(f"User {user.name} has no assigned quota.")
    logger.info(" ".join(more_remarks))  # log only reset message
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.post("/quota/{user_id}", response_model=LiveQuotaResp)
async def refresh_quota_policy_for_user(
    user_id: int = 0, name: Optional[str] = Body(None)
) -> LiveQuotaResp:
    """
    Accepts the id of the user whose quota policy should be refreshed.

    Returns the new remaining quota after the policy update in `response`
    """
    user, assoc_d, remarks = load_user_with_quota(user_id, name)
    quota = assoc_d[user_quota_assoc.assoc_name]
    oqp = OutboundQuotaPolicy()
    if not quota:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Cannot load quota for user {user.name}: none assigned.",
        )
    remarks.append(f"Quota policy config cache reset for {user.name}")
    response, more_remarks = oqp.refresh_policy_cache(user.name, quota)
    return LiveQuotaResp.send(response, remarks=remarks + more_remarks)


@api.post(
    "/config/write/",
    response_model=TextResp,
    responses={
        status.HTTP_401_UNAUTHORIZED: {
            "description": "Password does not match."
        }
    },
)
async def refresh_config_on_disk(passcode: str = Body(...)) -> TextResp:
    """Writes the current effective config to disk.

    Requires the CHAPPS password to be provided as the body.

    If line transmission security is an issue
      an SSL proxy layer will be
      required.  This is true for the entire application.

    """
    if (
        hash_password(passcode, config.chapps.payload_encoding)
        == config.chapps.password
    ):
        response = config.write()
        return TextResp.send(str(response))
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Password does not match.",
    )


@api.get("/sda/", response_model=SourceUserMapResp)
async def sda_batch_peek(
    user_ids: List[int], domain_ids: List[int] = [], email_ids: List[int] = []
) -> SourceUserMapResp:
    """Accepts `domain_ids`, `email_ids` and `user_ids` as body arguments:
    lists of integer object ids.

    Looks at current authorizations for all domain-user combinations.

    Returns their cache status as a dict of dicts, keyed as:
    `sda[domain-or-email][user] = SDAStatus`

    """
    sda = SenderDomainAuthPolicy()
    with Session() as sess:
        if domain_ids:
            domain_names = list(
                sess.scalars(Domain.select_names_by_id(domain_ids))
            )
        else:
            domain_names = []
        if email_ids:
            email_names = list(
                sess.scalars(Email.select_names_by_id(email_ids))
            )
        else:
            email_names = []
        if not domain_names and not email_names:
            raise HTTPException(
                status=status.HTTP_400_BAD_REQUEST,
                detail="No domains or emails were specified.",
            )
        user_names = list(sess.scalars(User.select_names_by_id(user_ids)))
    return SourceUserMapResp.send(
        sda.bulk_check_policy_cache(user_names, domain_names, email_names)
    )


@api.get("/sda/{source_name}/for/{user_name}", response_model=TextResp)
async def sda_peek(source_name: str, user_name: str) -> TextResp:
    """Accepts url-encoded domain name or email address and user name as path
    arguments.

    Returns status of cached SDA for the named user and source, i.e. is this
    user allowed to transmit email apparently from this domain or email address

    """
    sda = SenderDomainAuthPolicy()
    result = sda.check_policy_cache(user_name, source_name)
    # logger.debug(f"Peeking at {source_name} auth for {user_name}: {result!r}")
    return TextResp.send(result)


@api.delete("/sda/", response_model=TextResp)
async def sda_batch_clear(
    user_ids: List[int], domain_ids: List[int] = [], email_ids: List[int] = []
) -> TextResp:
    """
    Clears all source - user mappings by iterating through both lists:
    users and domains+emails.
    """
    sda = SenderDomainAuthPolicy()
    with Session() as sess:
        if domain_ids:
            domain_names = list(
                sess.scalars(Domain.select_names_by_id(domain_ids))
            )
        else:
            domain_names = []
        if email_ids:
            email_names = list(
                sess.scalars(Email.select_names_by_id(email_ids))
            )
        else:
            email_names = []
        if not domain_names and not email_names:
            raise HTTPException(
                status=status.HTTP_400_BAD_REQUEST,
                detail="No domains or emails specified.",
            )
        user_names = list(sess.scalars(User.select_names_by_id(user_ids)))
    sda.bulk_clear_policy_cache(user_names, domain_names, email_names)
    return TextResp.send(
        "SDA cache cleared for specified domains and/or emails x users."
    )


@api.delete("/sda/{source_name}/for/{user_name}", response_model=TextResp)
async def sda_clear(source_name: str, user_name: str) -> TextResp:
    """
    Accepts url-encoded domain name or email address and user name of SDA
    to clear.

    Returns the status of the SDA prior to clearing.
    """
    sda = SenderDomainAuthPolicy()
    return TextResp.send(sda.clear_policy_cache(user_name, source_name))


@api.get("/grl/tuple/", response_model=TimeResp)
async def grl_peek_tuple(
    client_address: str = Body(...),
    sender: str = Body(...),
    recipient: str = Body(...),
):
    """
    Accepts client IP address, sender email address and recipient email
    address as required arguments in the request body.

    Returns a float which is the UNIX epoch time of the last time that
    tuple was encountered by greylisting.
    """
    grl = GreylistingPolicy()
    try:
        timestamp = float(
            grl.redis.get(grl._tuple_key(client_address, sender, recipient))
        )
    except TypeError as e:
        return TimeResp.send(0.0)
    return TimeResp.send(timestamp)


@api.get("/grl/tally/{client_address}", response_model=InstanceTimesResp)
async def grl_list_tally(client_address: str):
    """Accepts the client IP address as the path argument.

    Returns a list of instance IDs and their timestamps as floats in UNIX epoch
    time (UTC).  In the standard time library, `localtime()` will convert them
    to a time struct in local time based on locale, or `gmtime()` will convert
    them to a struct in UTC.  Then `strftime()` may be used to format them for
    a human to read.

    """
    grl = GreylistingPolicy()
    tally = grl.redis.zrange(
        grl._client_key(client_address), 0, -1, withscores=True
    )
    tally_decoded = []
    if tally:
        tally_decoded = [(i.decode("utf-8"), float(t)) for i, t in tally]
    return InstanceTimesResp.send(tally_decoded)


@api.delete("/grl/tally/{client_address}", response_model=DeleteResp)
async def grl_clear_tally(client_address: str):
    """
    Accepts the client IP address as the path argument.

    Returns "deleted" if successful.
    """
    grl = GreylistingPolicy()
    grl.redis.delete(grl._client_key(client_address))
    return DeleteResp.send()


@api.delete("/grl/option_cache/{recipient_domain}", response_model=DeleteResp)
async def grl_clear_option_cache(recipient_domain: str):
    """
    Accepts a domain name as the path argument.

    Returns 'deleted' on success.
    """
    grl = GreylistingPolicy()
    grl.redis.delete(grl._domain_option_key(recipient_domain))
    return DeleteResp.send()


@api.delete("/spf/option_cache/{recipient_domain}", response_model=DeleteResp)
async def spf_clear_option_cache(recipient_domain: str):
    """
    Accepts a domain name as the path argument.

    Returns 'deleted' on success.
    """
    spf = SPFEnforcementPolicy()
    spf.redis.delete(spf._domain_option_key(recipient_domain))
    return DeleteResp.send()



"""
Policy Configuration Adapters
-----------------------------

Policy-configuration source data adapters.

All of these currently interface to MariaDB as a fairly low level.
A change is planned to convert all of this to use SQLAlchemy.

"""
import MySQLdb as dbmodule
import logging
from chapps.config import CHAPPSConfig

# from chapps.dbsession import sql_engine, sessionmaker
from chapps.models import User, Domain, Email, Quota
from contextlib import contextmanager
from typing import List, Dict, Union, Any

logger = logging.getLogger(__name__)  # pragma: no cover


class PolicyConfigAdapter:
    """Base class for policy config access

    Right now, all of the adapters use MariaDB, but this will change
    (as noted below).
    Non-database adapters could also be created, which might use a
    flat file, or a blockchain, or some sort of other API interaction,
    to be the source of policy data.

    Ideally there should be a second-level base class called something like
    `MariaDBPolicyConfigAdapter` which would hold all the SQL-specific stuff,
    but that would make a lot more sense if

      a) there were going to be a lot of weird, different adapters.  We may yet get there!
      b) there were anything else to put in the base class

    For now, however, this
    class serves to store instructions about how to create the **User** table,
    since they are used for everything, just about.

    .. admonition:: SQL constants are deprecated.

       All of the SQL statement constant strings defined in this class are
       deprecated.  In a future version, all of this logic will be accomplished
       using SQLAlchemy.

    """

    user_table = (
        "CREATE TABLE IF NOT EXISTS users ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(127) UNIQUE NOT NULL"
        ")"
    )
    """SQL query used to create the **user** table."""
    userid_query = "SELECT id FROM users WHERE name = %(user)s"
    """
    SQL query for obtaining the ID of a **user**.  These queries
    will go away when the older codebase is adapted to use SQLAlchemy
    """

    def __init__(
        self,
        *,
        cfg: CHAPPSConfig = None,
        db_host: str = None,
        db_port: int = None,
        db_name: str = None,
        db_user: str = None,
        db_pass: str = None,
        autocommit: bool = True,
    ):
        """
        :param str db_host: the hostname or IP address of the database server
        :param int db_port: the port number of the database server
        :param str db_name: the name of the database
        :param str db_user: the username for login
        :param str db_pass: the password for the user
        :param bool autocommit: defaults to True


        """
        self.config = cfg or CHAPPSConfig.get_config()
        self.params = self.config.adapter
        self.host = db_host or self.params.db_host or "127.0.0.1"
        self.port = db_port or self.params.db_port or 3306
        self.user = db_user or self.params.db_user
        self.pswd = db_pass or self.params.db_pass
        self.db = db_name or self.params.db_name
        self.autocommit = autocommit
        kwargs = dict(
            user=self.user,
            password=self.pswd,
            host=self.host,
            port=int(self.port),
            database=self.db,
            autocommit=self.autocommit,
        )
        self.conn = dbmodule.connect(**kwargs)

    def finalize(self):
        """Close the database connection."""
        self.conn.close()

    def _initialize_tables(self):
        """Set up required tables.

        Everything requires User records, so those tables are created here.

        """
        cur = self.conn.cursor()
        cur.execute(self.user_table)
        cur.close()

    @contextmanager
    def adapter_context(self):
        """Database connection context manager.

        This method is meant to be used as a context manager, like so:

        .. code:: python

          with someadapter.adapter_context() as cur:
              cur.execute(some_sql)

        The context manager closes the cursor once the context ends.
        """
        cur = self.conn.cursor()
        try:
            yield cur
        finally:
            cur.close()


class MariaDBQuotaAdapter(PolicyConfigAdapter):
    """An adapter for obtaining quota policy data from MariaDB

    .. admonition:: SQL constants are deprecated.

       All of the SQL statement constant strings defined in this class are
       deprecated.  In a future version, all of this logic will be accomplished
       using SQLAlchemy.

    """

    quota_table = (
        "CREATE TABLE IF NOT EXISTS quotas ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(31) UNIQUE NOT NULL,"
        "quota BIGINT UNIQUE NOT NULL"
        ")"
    )
    """SQL statement for creating the **Quota** table."""
    join_table = (
        "CREATE TABLE IF NOT EXISTS quota_user ("  # pragma: no cover
        "quota_id BIGINT NOT NULL,"
        "user_id BIGINT NOT NULL PRIMARY KEY,"
        "CONSTRAINT fk_user_quota"
        " FOREIGN KEY (user_id) REFERENCES users (id)"
        " ON DELETE CASCADE"
        " ON UPDATE RESTRICT,"
        "CONSTRAINT fk_quota"
        " FOREIGN KEY (quota_id) REFERENCES quotas (id)"
        " ON DELETE CASCADE"
        " ON UPDATE CASCADE"  # allow replacement of quota defs
        ")"
    )
    """SQL statement for creating the join table between **Quota** and **User** tables."""
    basic_quotas = (
        "INSERT INTO quotas ( name, quota ) VALUES "  # pragma: no cover
        "('10eph', 240),"
        "('50eph', 1200),"
        "('200eph', 4800)"
    )
    """Doubly deprecated insert statement for creating basic quotas."""
    quota_query = (
        "SELECT quota FROM quotas WHERE id = ("  # pragma: no cover
        "SELECT quota_id FROM quota_user AS j"
        " LEFT JOIN users AS u ON j.user_id = u.id"
        " WHERE u.name = %(user)s"
        ")"
    )
    """SQL query for selecting a **Quota**'s `quota` value for a named **User**."""
    quota_map_query = (
        "SELECT u.name AS user, p.quota FROM quotas AS p"  # pragma: no cover
        " LEFT JOIN quota_user AS j ON p.id = j.quota_id"
        " LEFT JOIN users AS u ON j.user_id = u.id"
    )
    """SQL query for selecting a list of users and their quotas."""
    quota_map_where = "WHERE u.name IN ({srch})"  # pragma: no cover
    r"""SQL for specifying which **User**\ s to returns quota values for."""

    def _initialize_tables(self, *, defquotas: bool = False):
        """Initialize tables required for this adapter/policy

        :py:mod:`chapps.policy.OutboundQuotaPolicy` requires a **Quota** object
        representing the user's outbound transmission limit.  The **Quota** has
        both a name and a value, for humans.

        """
        super()._initialize_tables()
        cur = self.conn.cursor()
        cur.execute(self.quota_table)
        cur.execute(self.join_table)
        if defquotas:
            cur.execute("SELECT COUNT(name) FROM quotas")
            if cur.fetchone()[0] == 0:
                cur.execute(self.basic_quotas)
        cur.close()

    def quota_for_user(self, user: str) -> Union[int, None]:
        """Return the quota amount for an user account

        :param str user: the user's name

        """
        cur = self.conn.cursor()
        cur.execute(self.quota_query, dict(user=user))
        try:
            res = cur.fetchone()[0]
        except TypeError:  # generally meaning no result; we could log this
            res = None
        except mariadb.Error as e:  # pragma: no cover
            logger.error(e)
            res = None
        finally:
            cur.close()
        return res

    def _quota_search(self, users: List[str] = None):
        """Return selected rows of username and quota-amount

        Use deprecated stored constant SQL fragments to search for the
        specified users, or select all users.

        Return value is similar to:

        .. code::python

          [ ("username1", 200), ("username2", 400) ]

        """
        users = users or []
        cur = self.conn.cursor()
        if len(users) == 0:
            query = self.quota_map_query
            cur.execute(query)
        else:
            query = self.quota_map_query + " " + self.quota_map_where
            srch = str(users)[
                1:-1
            ]  # produces a string suitable for SQL "IN ()"
            cur.execute(query.format(srch=srch))
        return cur.fetchall()

    def quota_dict(self, users: List[str] = None) -> Dict[str, str]:
        """Return a dict which maps users onto their quotas

        :param List[str] users: Optional limiting list of usernames; if
          empty or not provided, all users will be listed.

        """
        users = users or []
        rows = self._quota_search(users)
        res = {r[0]: r[1] for r in rows}
        return res

    def quota_map(self, func: callable, users: List[str] = None) -> List[Any]:
        """Map a callable over a set of users and their quotas

        :param callable func: A callable which will be passed (username,
           quota amount) for each user in the list

        :param List[str] users: Optional limiting list of usernames; if
          empty or not provided, all users will be listed.

        Provide a function to execute over each user and its quota.  Use this
        to directly wire the database-loading logic to the Redis-population
        logic.

        """
        users = users or []
        if not callable(func):
            raise ValueError(
                "The first non-self argument must be a callable which accepts the user and quota as arguments, in that order."
            )
        rows = self._quota_search(users)
        res = []
        for row in rows:
            res.append(func(row[0], row[1]))
        return res


class MariaDBSenderDomainAuthAdapter(PolicyConfigAdapter):
    """An adapter to obtain sender domain authorization data from MariaDB

    .. admonition:: SQL constants are deprecated.

       All of the SQL statement constant strings defined in this class are
       deprecated.  In a future version, all of this logic will be accomplished
       using SQLAlchemy.

"""

    domain_table = (
        "CREATE TABLE IF NOT EXISTS domains ("  # pragma: no cover
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(63) UNIQUE NOT NULL,"
        "greylist TINYINT(1) NOT NULL DEFAULT 0,"
        "check_spf TINYINT(1) NOT NULL DEFAULT 0"
        ")"
    )
    domain_join_table = (
        "CREATE TABLE IF NOT EXISTS domain_user ("  # pragma: no cover
        "domain_id BIGINT NOT NULL,"
        "user_id BIGINT NOT NULL,"
        "PRIMARY KEY (domain_id, user_id),"  # comp. primary key allows more than one user per domain
        "CONSTRAINT fk_user_domain"
        " FOREIGN KEY (user_id) REFERENCES users (id)"
        " ON DELETE CASCADE"
        " ON UPDATE RESTRICT,"
        "CONSTRAINT fk_domain"
        " FOREIGN KEY (domain_id) REFERENCES domains (id)"
        " ON DELETE CASCADE"
        " ON UPDATE CASCADE"  # allow replacement of domain defs
        ")"
    )
    email_table = (
        "CREATE TABLE IF NOT EXISTS emails ("
        "id BIGINT AUTO_INCREMENT PRIMARY KEY,"
        "name VARCHAR(127) UNIQUE NOT NULL"
        ")"
    )
    email_join_table = (
        "CREATE TABLE IF NOT EXISTS email_user ("
        "email_id BIGINT NOT NULL,"
        "user_id BIGINT NOT NULL,"
        "PRIMARY KEY (email_id, user_id),"
        "CONSTRAINT fk_user_email"
        " FOREIGN KEY (user_id) REFERENCES users (id)"
        " ON DELETE CASCADE"
        " ON UPDATE RESTRICT,"
        "CONSTRAINT fk_email"
        " FOREIGN KEY (email_id) REFERENCES emails (id)"
        " ON DELETE CASCADE"
        " ON UPDATE CASCADE"
        ")"
    )
    domain_query = (
        "SELECT d.name FROM domains AS d"  # pragma: no cover
        " WHERE id = ("
        "SELECT domain_id FROM domain_user AS j"
        " LEFT JOIN users AS u ON j.user_id = u.id"
        " WHERE u.name = %(user)s"
        ")"
    )
    domain_map_query = (
        "SELECT u.name AS user, d.name AS domain FROM domains AS d"  # pragma: no cover
        " LEFT JOIN domain_user AS j ON p.id = j.domain_id"
        " LEFT JOIN users AS u ON j.user_id = u.id"
    )
    domain_map_where = "WHERE user IN ({srch})"  # pragma: no cover
    check_domain_query = (
        "SELECT COUNT(d.name) FROM domains AS d"
        " LEFT JOIN domain_user AS j ON d.id = j.domain_id"
        " LEFT JOIN users AS u ON u.id = j.user_id"
        " WHERE d.name = '{domain}' AND u.name = '{user}'"
    )
    check_email_query = (
        "SELECT COUNT(e.name) FROM emails AS e"
        " LEFT JOIN email_user AS j on e.id = j.email_id"
        " LEFT JOIN users AS u ON u.id = j.user_id"
        " WHERE e.name = '{email}' AND u.name = '{user}'"
    )

    def _initialize_tables(self, *args, **kwargs):
        """Initialize tables required for this adapter/policy

        :param List args: catch all positional arguments
        :param Dict kwargs: catch all keyword arguments

        Sender domain authorization also checks whole-email matches
        as a fallback, and so the policy requires storage for domain
        objects and also email objects.

        Does not currently make use of any arguments.
        """
        super()._initialize_tables()
        cur = self.conn.cursor()
        cur.execute(self.domain_table)
        cur.execute(self.email_table)
        cur.execute(self.domain_join_table)
        cur.execute(self.email_join_table)
        cur.close()

    def check_domain_for_user(self, user: str, domain: str) -> bool:
        """Returns True if the user is authorized to send for this domain

        :param str user: name of user
        :param str domain: name of domain

        """
        cur = self.conn.cursor()
        cur.execute(self.check_domain_query.format(user=user, domain=domain))
        result = cur.fetchone()[0]  ### returns 1 if a domain matched, 0 if not
        cur.close()
        return result

    def check_email_for_user(self, user: str, email: str) -> bool:
        """Returns True if the user is authorized to send as this email

        :param str user: name of user
        :param str email: email address

        """
        cur = self.conn.cursor()
        cur.execute(self.check_email_query.format(user=user, email=email))
        result = cur.fetchone()[0]
        cur.close()
        return result


class MariaDBInboundFlagsAdapter(PolicyConfigAdapter):
    """Adapter class for handling option flags for inbound mail

    Generally these flags are set on a per-domain basis.
    """

    greylist_on_domain_query = (
        "SELECT greylist FROM domains WHERE name = '{domain}' LIMIT 1"
    )

    check_spf_on_domain_query = (
        "SELECT check_spf FROM domains WHERE name = '{domain}' LIMIT 1"
    )

    def do_greylisting_on(self, domain: str) -> bool:
        cur = self.conn.cursor()
        cur.execute(self.greylist_on_domain_query.format(domain=domain))
        try:
            result = cur.fetchone()[0]
        except TypeError:
            result = 0
        finally:
            cur.close()
        return result == 1

    def check_spf_on(self, domain: str) -> bool:
        cur = self.conn.cursor()
        cur.execute(self.check_spf_on_domain_query.format(domain=domain))
        try:
            result = cur.fetchone()[0]
        except TypeError:
            result = 0
        finally:
            cur.close()
        return result == 1



"""
Signal handlers and custom exceptions
-------------------------------------

In order to shut down gracefully, we want to provide a signal handler when a
service (handler) loop starts.  Also, CHAPPS raises a few different custom
exceptions, which are defined here.

"""
import signal
import logging

logger = logging.getLogger(__name__)


class SignalHandlerFactory:  # pragma: no cover
    """A class for containing classmethods to create signal handlers"""

    # could be organized differently; instantiating the class would return
    # an instance with __call__ defined, so that it is callable, and then
    # would in turn call the closure function;

    @classmethod
    def signal_handler(cls, loop=None) -> callable:
        """Returns a signal-checking, exiting closure

        :param loop: *Deprecated*

        :returns: a closure that looks for :const:`~signal.SIGTERM` or
          :const:`~signal.SIGINT` and raises :exc:`SystemExit` if it finds
          either one.

        .. todo::

          Since the returned closure is actually completely static, this could
          be simplified greatly.  However, it works and doesn't cost much, so
          it low on the list.

        """

        def signal_handler_closure(sig, frame=None):
            if sig in {signal.SIGTERM, sig.SIGINT}:
                logger.info(
                    f"CHAPPS exiting on {signal.Signals(sig)} ({sig})."
                )
                raise SystemExit

        return signal_handler_closure


class CHAPPSException(Exception):
    """Parent class for CHAPPS exceptions"""


class CallableExhausted(CHAPPSException):
    """A special exception for use during testing"""


class NotAnEmailAddressException(CHAPPSException):
    """An email address contained no at-signs"""


class TooManyAtsException(CHAPPSException):
    """An email address had too many at-signs in it"""


class ConfigurationError(CHAPPSException):
    """There was an error in the setting of configuration elements"""


class OutboundPolicyException(CHAPPSException):
    """Parent of exceptions which occur during outbound mail processing"""


class NullSenderException(OutboundPolicyException):
    """No sender address exists in the current policy request"""


class AuthenticationFailureException(OutboundPolicyException):
    """Lack of user-identifier being treated as authentication failure"""


class InboundPolicyException(CHAPPSException):
    """Parent of exceptions arising during inbound email processing"""


class NoRecipientsException(InboundPolicyException):
    """This is raised if the recipient field is somehow empty"""


class NoSuchDomainException(CHAPPSException):
    """A domain operation was attempted against a nonexistent Domain"""


class HELOWLException(CHAPPSException):
    """HELO Whitelisting exceptions"""


class AddressDoesNotMatchDNS(HELOWLException):
    """A supplied IP address in the HELO Whitelist doesn't match the A record"""


class NameDoesNotMatchPTR(HELOWLException):
    """The IP address obtained for the HELO name has a PTR record which does
    not match the name supplied"""



"""
CHAPPS data schemata
--------------------

Database schema expressed as `SQLAlchemy`_ `ORM Models`_

Basic data models correspond to Pydantic models for API, but more models are
required for the database side:

  * There is a subclass of the metaclass,
    :class:`~sqlalchemy.orm.DeclarativeMeta`, which provides extra logic to
    the database models to automate common queries and tasks.

  * There are table definitions for the join tables.

  * There is a special class for generalizing the logic of joining the tables.

.. todo::

    it seems like there must be some way to obtain more of the data
    tracked by :class:`.JoinAssoc` from the :mod:`SQLAlchemy` metadata but I
    have not figured much of it out yet.)

.. _orm models: https://docs.sqlalchemy.org/en/14/orm/quickstart.html

"""
from typing import Dict, List, Union, Optional, Any
from sqlalchemy import (
    Column,
    Integer,
    String,
    Boolean,
    ForeignKey,
    Table,
    select,
    update,
    tuple_,
    delete,
)
from sqlalchemy.orm import (
    declarative_base,
    relationship,
    backref,
    DeclarativeMeta,
    selectinload,
)
from sqlalchemy.schema import MetaData
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",  # or column_0_name
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
"""SQLA's recommended naming convention for constraints"""

# declare subclass of the SQLAlchemy DeclarativeMeta class
# in order to attach custom routines to the ORM objects
class DB_Customizations(DeclarativeMeta):
    """Custom ORM metaclass

    In order to make common tasks more concise, and to conceal the need for
    complex data massaging in order to produce appropriate data structures for
    the different kinds of queries, a fair amount of custom code is introduced
    into the ORM models via this metaclass.  These routines allow calls to the
    ORM class itself, to get it to return statement objects (objects defined in
    :mod:`sqlalchemy.sql.expression`).

    """

    def select_by_id(cls, id: int):
        """SELECT (load) a single object by ID"""
        return select(cls).where(cls.id == id)

    def windowed_list_by_ids(
        cls,
        *,
        ids: List[int] = [],
        subquery=None,
        skip: int = 0,
        limit: int = 1000,
        q: str = "%",
    ):
        """SELECT a window of objects

        :param Optional[List[int]] ids: a list of IDs
        :param Optional[~sqlalchemy.sql.expression.Subquery] subquery: a
          sub-SELECT which yields IDs
        :param Optional[int] skip: start the window after `skip` entries
        :param Optional[int] limit: include up to `limit` entries in the
          result
        :param Optional[str] q: a substring to match against the `name` field

        :raises ValueError: unless one of `ids` or `subquery` is provided

        :returns: a :class:`~sqlalchemy.sql.expression.Select` implementing the window

        :rtype: sqlalchemy.sql.expression.Select

        If both of `subquery` and `ids` are not empty, `subquery`
        prevails and `ids` is ignored.

        """
        if subquery is not None:
            stmt = select(cls).where(cls.id.in_(subquery))
        elif ids:
            stmt = select(cls).where(tuple_(cls.id).in_([(i,) for i in ids]))
        else:
            raise ValueError("Supply one of ids or subquery.")
        stmt = (
            stmt.where(cls.name.like(q))
            .offset(skip)
            .limit(limit)
            .order_by(cls.id)
        )
        return stmt

    def select_by_ids(cls, ids: List[int], assoc: Optional[Any] = None):
        """Return a select statement for a list of objects,
           optionally with eager-loaded associations
        """
        stmt = select(cls).where(tuple_(cls.id).in_([(i,) for i in ids]))
        if assoc:
            stmt = stmt.options(selectinload(assoc))
        stmt = stmt.order_by(cls.id)
        return stmt

    def select_names_by_id(cls, ids: List[int]):
        """Return a Select for the names corresponding to the provided IDs"""
        return select(cls.name).where(tuple_(cls.id).in_([(i,) for i in ids]))

    def select_by_pattern(cls, q: str):
        """Return a Select for all records with names which include `q` as a substring"""
        return select(cls).where(cls.name.like(q))

    def select_by_name(cls, q: str):
        """Return a Select for the record whose name exactly matches `q`"""
        return select(cls).where(cls.name == q)

    def windowed_list(cls, q: str = "%", skip: int = 0, limit: int = 1000):
        """Return a Select for a window of :meth:`.select_by_pattern`"""
        return (
            cls.select_by_pattern(q).offset(skip).limit(limit).order_by(cls.id)
        )

    def remove_by_id(
        cls, ids: Union[int, List[int]]
    ):  # (i,) creates a tuple w/ 1 element
        """Return a Delete for the listed IDs (`ids` may be a scalar ID also)"""
        if type(ids) == int:
            ids = [ids]
        return delete(cls).where(tuple_(cls.id).in_([(i,) for i in ids]))

    def update_by_id(cls, item):
        """Return an Update statement for the specified item

        :param type cls: an ORM model class (subclass of :const:`~.DB_Base`)
        :param cls item: an instance of an ORM model of the type stored in `cls`

        :returns: an :class:`~sqlalchemy.sql.expression.Update` statement object representing the new values of `item`

        :rtype: sqlalchemy.sql.expression.Update

        """
        args = {
            k: getattr(item, k)
            for k in item.schema()["properties"].keys()
            if getattr(item, k, None) is not None
        }
        id = args.pop("id")
        return update(cls).where(cls.id == id).values(**args)


# declare DB model base class
DB_Base = declarative_base(
    metaclass=DB_Customizations,
    metadata=MetaData(naming_convention=convention),
)
"""DB_Base serves as the base of all `ORM models`_

This class itself contains literally no code apart from documentation.
All of the magic provided for the ORM layer is implemented in the metaclass,
:class:`~.DB_Customizations`.  To be clear, all of the methods defined in
the metaclass become available as class methods on the classes derived from
:class:`~.DB_Base`, because it is of type :class:`.DB_Customizations`.

"""


quota_user = Table(
    "quota_user",
    DB_Base.metadata,
    Column(
        "user_id",
        ForeignKey("users.id", ondelete="CASCADE", onupdate="RESTRICT"),
        primary_key=True,
    ),
    Column(
        "quota_id",
        ForeignKey("quotas.id", ondelete="CASCADE", onupdate="CASCADE"),
        nullable=False,
    ),
)
"""the `quota_user` join table"""

domain_user = Table(
    "domain_user",
    DB_Base.metadata,
    Column(
        "user_id",
        ForeignKey("users.id", ondelete="CASCADE", onupdate="RESTRICT"),
        primary_key=True,
    ),
    Column(
        "domain_id",
        ForeignKey("domains.id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    ),
)
"""the `domain_user` join table"""

email_user = Table(
    "email_user",
    DB_Base.metadata,
    Column(
        "user_id",
        ForeignKey("users.id", ondelete="CASCADE", onupdate="RESTRICT"),
        primary_key=True,
    ),
    Column(
        "email_id",
        ForeignKey("emails.id", ondelete="CASCADE", onupdate="CASCADE"),
        primary_key=True,
    ),
)
"""the `email_user` join table"""


class Quota(DB_Base):
    """ORM Model for **Quota** definitions

    Each record/instance has these columns/attributes:

    """

    __tablename__ = "quotas"

    id = Column(Integer, primary_key=True)
    """integer primary key"""
    name = Column(String(32), unique=True, nullable=False, index=True)
    """unique string of 32 chars or less"""
    quota = Column(Integer, unique=True, nullable=False, index=True)
    """unique integer transmission attempt limit"""

    def __repr__(self):
        return f"Quota[ORM](id={self.id!r}, name={self.name!r}, quota={self.quota!r})"


class Domain(DB_Base):
    """ORM Model for **Domain** definitions

    Each record/instance has these columns/attributes:

    """

    __tablename__ = "domains"

    id = Column(Integer, primary_key=True)
    """integer primary key"""
    name = Column(String(64), unique=True, nullable=False, index=True)
    """unique string of 64 chars or less"""
    greylist = Column(
        Boolean(name="greylist"),
        nullable=False,
        server_default="0",
        default=False,
    )
    """if True perform greylisting"""
    check_spf = Column(
        Boolean(name="check_spf"),
        nullable=False,
        server_default="0",
        default=False,
    )
    """if True perform SPF enforcement"""

    def __repr__(self):
        return (
            f"Domain[ORM](id={self.id!r}, "
            f"name={self.name!r}, "
            f"greylist={'Y' if self.greylist else 'N'}, "
            f"check_spf={'Y' if self.check_spf else 'N'})"
        )


class Email(DB_Base):
    """ORM Model for **Email** definitions

    Each record/instance has these columns/attributes:

    """

    __tablename__ = "emails"

    id = Column(Integer, primary_key=True)
    """integer primary key"""
    name = Column(String(128), unique=True, nullable=False, index=True)
    """unique string of 128 chars or less"""

    def __repr__(self):
        return f"Email[ORM](id={self.id!r}, name={self.name!r})"


class User(DB_Base):
    """ORM model for **User** objects

    Each record/instance has these columns/attributes:

    """

    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    """integer auto-incremented primary key"""
    name = Column(String(128), unique=True, nullable=False, index=True)
    """unique string of up to 128 chars"""
    quota = relationship(
        Quota,
        secondary=quota_user,  # using m2m mainly for optimization
        backref="users",  # wants to be efficient and non-eager
        passive_deletes=True,  # prevent loading during deletion
        single_parent=True,  # only one association to quota
        uselist=False,  # one quota per user
        primaryjoin=id == quota_user.c.user_id,
        secondaryjoin=Quota.id == quota_user.c.quota_id,
    )
    """associated **Quota** object; there can be only one **Quota** associated to a **User**"""
    domains = relationship(
        Domain,
        secondary=domain_user,  # really m2m
        backref=backref("users", order_by="User.id"),  # reverse associate
        order_by=Domain.id,  # order by id
        passive_deletes=True,  # protect domains
    )
    """list of associated **Domain** objects

    a **User** may be associated to more than one **Domain**, and
    a **Domain** may be associated to more than one **User**

    """
    emails = relationship(
        Email,
        secondary=email_user,
        backref=backref("users", order_by="User.id"),
        order_by=Email.id,
        passive_deletes=True,
    )
    """list of associated **Email** objects

    a **User** may be associated to more than one **Email**, and
    an **Email** may be associated to more than one **User**

    """

    def __repr__(self):
        return f"User[ORM](id={self.id!r}, name={self.name!r})"


class JoinAssoc:
    """Represent joining two data tables via a two-column join table

    There is probably some way to use SQLA for this but I am using it wrongly,
    to avoid loading associations just in order to link them to other objects
    via a join table

    """

    def __init__(
        self,
        source_model: DB_Base,
        source_id: str,
        *,
        assoc_name: str,
        assoc_type: type,
        assoc_model: DB_Base,
        assoc_id: str,
        table,
    ):
        """Define a join association

        :param DB_Base source_model: a reference to the dbmodel class of the
          source object

        :param str source_id: label of the source object's ID column in the
          join table

        :param str assoc_name: attribute name of the association

        :param type assoc_type: usually :obj:`int` or :obj:`List[int]`; should
          be the type for the API to expect when setting up the route metadata

        :param DB_Base assoc_model: a reference to the dbmodel class of the
          associated object

        :param str assoc_id: label of associated object's ID column in the
          join table

        :param sqlalchemy.schema.Table table: a reference to the join table
          schema; it will be a constant in this module, generally

        """
        self.source_model = source_model
        self.source_id = source_id
        self.assoc_name = assoc_name
        self.assoc_type = assoc_type
        self.assoc_model = assoc_model
        self.assoc_id = assoc_id
        self.table = table

    def __repr__(self):
        return (
            f"JoinAssoc({self.source_model.__name__}, "
            f"{self.assoc_model.__name__}, "
            f"{self.assoc_name}, {self.assoc_type!r})"
        )

    def select(self):
        """Convenience method

        :returns: a SELECT on the join table

        :rtype: sqlalchemy.sql.expression.Select

        """
        return self.table.select()

    def insert(self):
        """Convenience method

        :returns: an INSERT on the join table

        :rtype: sqlalchemy.sql.expression.Insert

        """
        return self.table.insert()

    def delete(self):
        """Convenience method

        :returns: a DELETE FROM on the join table

        :rtype: sqlalchemy.sql.expression.Delete

        """
        return self.table.delete()

    def update(self):
        """Convenience method

        :returns: an UPDATE on the join table

        :rtype: sqlalchemy.sql.expression.Update

        """
        return self.table.update()

    def values(
        self, item, assoc: Union[int, List[int]]
    ) -> List[Dict[str, int]]:
        """Get value-insertion or update mapping

        :param Union[DB_Base, int] item: the source item or its ID

        :param Union[int,List[int]] assoc: the association or associations to
          specify

        Using the `item` passed in, which may be an ORM object descended from
        :class:`~.DB_Base` or an :obj:`int`, construct a list of dictionaries
        mapping the join table's ID column names correctly onto the source item
        ID and the associated item ID(s).

        """

        item_id = item if type(item) == int else item.id
        try:
            i = iter(assoc)
            ins = [{self.source_id: item_id, self.assoc_id: val} for val in i]
        except TypeError:
            ins = [{self.source_id: item_id, self.assoc_id: assoc}]
        return ins

    def where_tuples(self, item_id: int, assoc: Union[int, List[int]]):
        """Get tuples suitable for use in an SQLAlchemy WHERE clause"""
        try:
            i = iter(assoc)
            res = [(item_id, val) for val in i]
        except TypeError:
            res = [(item_id, assoc)]
        return res

    @property
    def source_col(self):
        return getattr(self.table.c, self.source_id)

    @property
    def assoc_col(self):
        return getattr(self.table.c, self.assoc_id)

    def insert_assoc(self, item_id: int, vals):
        return (
            self.insert()
            .prefix_with("IGNORE")
            .values(self.values(item_id, vals))
        )

    def delete_assoc(self, item_id: int, vals):
        return self.delete().where(
            tuple_(self.source_col, self.assoc_col).in_(
                self.where_tuples(item_id, vals)
            )
        )

    def update_assoc(self, item_id: int, assoc_id: int):
        return (
            self.update()
            .where(self.source_col == item_id)
            .values(**{self.assoc_id: assoc_id})
        )

    def select_ids_by_source_id(self, item_id: int):
        # not self.select() b/c we want the SQLA select()
        return select(self.assoc_col).where(self.source_col == item_id)



"""CHAPPS

Caching, Highly-Available Postfix Policy Service

In addition to module documentation found here, an introduction and overview of
the project is presented in its README_.  Specific concerns and instructions
about getting it running are discussed in the INSTALLATION_ file.  For a
history of changes to the project, see the CHANGELOG_.

"""
from ._version import __version__
import chapps.logging

__all__ = [
    "_version",
    "util",
    "config",
    "signals",
    "logging",
    "sqla_adapter",
    "policy",
    "spf_policy",
    "inbound",
    "outbound",
    "switchboard",
    "dbmodels",
    "models",
    "dbsession",
]



"""
Policy Configuration Adapters based on SQLAlchemy
-------------------------------------------------

Policy-configuration source data adapters.

These adapter classes have been adjusted from their original form to use
SQLAlchemy.

"""
import logging
from chapps.config import CHAPPSConfig
from chapps.dbsession import (
    create_db_url,
    create_engine,
    sql_engine,
    sessionmaker,
    func,
    select,
)
from chapps.signals import NoSuchDomainException
from chapps.models import User, Domain, Email, Quota
from chapps import dbmodels
from contextlib import contextmanager
from typing import List, Dict, Union, Any

logger = logging.getLogger(__name__)  # pragma: no cover


class SQLAPolicyConfigAdapter:
    """Base class for policy config access using SQLAlchemy_"""

    def __init__(
        self,
        *,
        cfg: CHAPPSConfig = None,
        db_host: str = None,
        db_port: int = None,
        db_name: str = None,
        db_user: str = None,
        db_pass: str = None,
        autocommit: bool = True,
    ):
        """
        :param str db_host: the hostname or IP address of the database server
        :param int db_port: the port number of the database server
        :param str db_name: the name of the database
        :param str db_user: the username for login
        :param str db_pass: the password for the user
        :param bool autocommit: defaults to True


        """
        self.config = cfg or CHAPPSConfig.get_config()
        self.params = self.config.adapter
        # specifically: use the global engine unless we were passed a config
        # logger.debug("Using config file: " + config.chapps.config_file)
        if cfg:
            logger.debug(
                "Passed override config based on " + cfg.chapps.config_file
            )
        # logger.debug("Global sql_engine is " + str(sql_engine))
        self.sql_engine = (
            create_engine(create_db_url(cfg)) if cfg else sql_engine
        )
        # logger.debug("Using sql_engine " + str(self.sql_engine))

    def finalize(self):
        """Do nothing.  A no-op to maintain backward compatibility."""
        pass

    def _initialize_tables(self):
        """Set up required tables.

        The schemata for the tables are defined in :mod:`~.dbmodels`.  We can
        create everything in a stroke by simply invoking :func:`create_all` on
        the SQLAlchemy_ metadata_.

        .. _metadata: https://docs.sqlalchemy.org/en/14/tutorial/metadata.html#id1

        """
        dbmodels.DB_Base.metadata.create_all(self.sql_engine)


class SQLAQuotaAdapter(SQLAPolicyConfigAdapter):
    """An adapter for obtaining quota policy data from MariaDB
       using SQLAlchemy_

    """

    def quota_for_user(self, user: str) -> Union[int, None]:
        """Return the quota amount for an user account

        :param str user: the user's name

        """
        Session = sessionmaker(self.sql_engine)
        with Session() as sess:
            try:
                u = sess.execute(User.select_by_name(user)).scalar()
                return u.quota.quota
            except AttributeError:
                return None


class SQLASenderDomainAuthAdapter(SQLAPolicyConfigAdapter):
    """An adapter to obtain sender domain authorization data from MariaDB"""

    def check_domain_for_user(self, user: str, domain: str) -> bool:
        """Returns True if the user is authorized to send for this domain

        :param str user: name of user
        :param str domain: name of domain

        """
        Session = sessionmaker(self.sql_engine)
        with Session() as sess:
            user_subselect = (
                select(User.id).where(User.name == user).scalar_subquery()
            )
            domain_subselect = (
                select(Domain.id)
                .where(Domain.name == domain)
                .scalar_subquery()
            )
            stmt = (
                select(dbmodels.domain_user)
                .where(dbmodels.domain_user.c.user_id == user_subselect)
                .where(dbmodels.domain_user.c.domain_id == domain_subselect)
            )
            res = sess.execute(stmt)
            return len(list(res.scalars()))

    def check_email_for_user(self, user: str, email: str) -> bool:
        """Returns True if the user is authorized to send as this email

        :param str user: name of user
        :param str email: email address

        """
        Session = sessionmaker(self.sql_engine)
        with Session() as sess:
            user_subselect = (
                select(User.id).where(User.name == user).scalar_subquery()
            )
            email_subselect = (
                select(Email.id).where(Email.name == email).scalar_subquery()
            )
            stmt = (
                select(dbmodels.email_user)
                .where(dbmodels.email_user.c.user_id == user_subselect)
                .where(dbmodels.email_user.c.email_id == email_subselect)
            )
            res = sess.execute(stmt)
            return len(list(res.scalars()))


class SQLAInboundFlagsAdapter(SQLAPolicyConfigAdapter):
    def do_greylisting_on(self, domain: str):
        """Returns true if the domain enforces greylisting, otherwise False

        :param domain: full domain part

        :raises NoSuchDomainException: if the domain doesn't exist

        """
        Session = sessionmaker(self.sql_engine)
        with Session() as sess:
            domain_select = Domain.select_by_name(domain)
            domain_record = sess.execute(domain_select).scalar()
            if not domain_record:
                raise NoSuchDomainException(domain)
            return domain_record.greylist

    def check_spf_on(self, domain: str):
        """Returns true if the domain enforces SPF policies, else False

        :param domain: full domain part

        :raises NoSuchDomainException: if the domain doesn't exist

        """
        Session = sessionmaker(self.sql_engine)
        with Session() as sess:
            domain_select = Domain.select_by_name(domain)
            domain_record = sess.execute(domain_select).scalar()
            if not domain_record:
                raise NoSuchDomainException(domain)
            return domain_record.check_spf



"""
SQLAlchemy DB session global setup
----------------------------------

This module's purpose is to present a simple and unencumbered module to include
in order to access the database according to the configured credentials.

"""
from sqlalchemy import create_engine, func, select
from sqlalchemy.engine import URL
from sqlalchemy.orm import sessionmaker
from chapps.config import CHAPPSConfig
import logging

logger = logging.getLogger(__name__)

DIALECT_MAP = dict(mariadb="mysql", mysql="mysql")
"""Map dialects to drivers, for DBI URL construction"""


def create_db_url(cfg: CHAPPSConfig = None) -> URL:
    """Create a DBI URL for initializing :mod:`SQLAlchemy`

    :param cfg: optional config override

    :returns: `URL <https://docs.sqlalchemy.org/en/14/core/engines.html#sqlalchemy.engine.URL>`_ instance for use in accessing the database

    """
    cfg = cfg or CHAPPSConfig.get_config()
    if not cfg.venvdetector.sb:
        logger.debug("Using config file: " + cfg.chapps.config_file)
    adapter = cfg.adapter
    if adapter.adapter not in DIALECT_MAP:
        raise ValueError(
            (
                "Configured database adapter must be one of:"
                f"{', '.join([repr(v) for v in DIALECT_MAP.keys()])}"
            )
        )
    dialect = DIALECT_MAP[adapter.adapter]
    creds = dict(
        password=adapter.db_pass,  # auto encoded
        username=adapter.db_user,
        host=adapter.db_host or "127.0.0.1",
        port=adapter.db_port or "3306",
        database=adapter.db_name,
    )
    return URL.create(dialect, **creds)


sql_engine = create_engine(create_db_url())
"""a package-global SQL engine"""



"""
Top level API definition
------------------------

This module includes the other API-related modules
and configures the main :mod:`FastAPI` object which
answers web requests.  That object's name is `api`,
making its absolute symbol path :const:`chapps.rest.api.api`
"""
# Note that the API is powered by FastAPI and as such, the main API object
# itself is designed to be executed by uvicorn, probably a uvicorn worker
# coordinated by gunicorn (see FastAPI deployment docs)

from chapps.config import CHAPPSConfig
from chapps._version import __version__
from chapps.rest.routers import users, domains, quotas, emails, live
from fastapi import FastAPI, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
import logging
import chapps.logging
from pathlib import Path

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

config = CHAPPSConfig.get_config()

rest_readme = Path(config.chapps.docpath) / "README-API.md"
# Use the VenvDetector-derived docpath to locate the API readme

desc = rest_readme.open("rt").read()
# Read in the contents of the API readme

tags_metadata = [
    dict(
        name="users",
        description=(
            "<h3>Create, list, fetch, update, and delete operations "
            "involving users.</h3><p>Use the User Replace Quota route "
            "to assign a new <b>quota</b>.  There are some special routes "
            "for listing, allowing, or denying just some <b>domains</b>"
            " or <b>emails</b> without "
            "having to handle the entire set of associations.</p>"
        ),
    ),
    dict(
        name="domains",
        description=(
            "<h3>Create, list, fetch, update and delete operations for "
            "domains.</h3><p>When creating a <b>domain</b>, "
            "it is possible to specify a list of <b>user</b> IDs to "
            "associate.  There are some special routes for managing"
            " <b>user</b> associations without having"
            " to handle the entire set of associations.</p>"
        ),
    ),
    dict(
        name="quotas",
        description=(
            "<h3>Create, list, fetch, update and delete operations for "
            "quotas.</h3><p>Since there are very "
            "few <b>quotas</b> compared to the number of <b>users</b>, there is "
            "no support planned for managing <b>users</b> from their "
            "associated <b>quotas</b>.  To change a <b>quota</b> for a <b>user</b>, "
            "make use of the User Replace Quota route.</p>"
        ),
    ),
    dict(
        name="emails",
        description=(
            "<h3>Create, list, fetch, update and delete operations for "
            "email address objects.</h3>"
            "<p>When creating an <b>email</b>, it is possible to specify "
            "a list of <b>user</b> IDs to associate.  There are some special"
            " routes for managing <b>user</b> associations without having to"
            " handle the entire set of associations.</p>"
        ),
    ),
    dict(
        name="live",
        description=(
            "<h3>Status reporting and remote command interface.</h3><p>Routes "
            "provided for obtaining real-time remaining quota, resetting "
            "quota, refreshing policy settings, adding or removing SDAs, "
            "and causing CHAPPS to rewrite its config file, in order to "
            "allow for updates post-upgrade without losing site customizations.</p>"
        ),
    ),
]

api = FastAPI(
    title="CHAPPS REST API",
    description=desc,
    version=__version__,
    contact=dict(
        name="Caleb S. Cullen",
        url="https://github.com/easydns/chapps",
        email="ccullen@easydns.com",
    ),
    license_info=dict(name="MIT License", url="https://mit-license.org/"),
    openapi_tags=tags_metadata,
)
# The top-level :class:`fastapi.FastAPI` object

# verstr = config.chapps.version
api.include_router(users.api)
api.include_router(domains.api)
api.include_router(quotas.api)
api.include_router(emails.api)
api.include_router(live.api)


@api.exception_handler(RequestValidationError)
async def validation_exception_handler(
    request: Request, exc: RequestValidationError
):
    """Creates a log entry upon request validation error

    Also returns a reasonable response with code 422 to the browser.

    """
    exc_str = f"{exc}".replace("\n", " ").replace("   ", " ")
    body = await request.json()
    logging.error(f"{body!r}: {exc_str}")
    content = {"status_code": 10422, "message": exc_str, "data": None}
    return JSONResponse(
        content=content, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
    )



"""adds SPF and Greylisting defaulting to non-enforcement

Revision ID: 0b5ef72836bd
Revises: ce81767cec26
Create Date: 2022-06-20 13:54:10.920850

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0b5ef72836bd"
down_revision = "ce81767cec26"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "domains",
        sa.Column(
            "greylist",
            sa.Boolean(name="greylist"),
            nullable=False,
            server_default="0",
        ),
    )
    op.add_column(
        "domains",
        sa.Column(
            "check_spf",
            sa.Boolean(name="check_spf"),
            nullable=False,
            server_default="0",
        ),
    )


def downgrade() -> None:
    op.drop_column("domains", "check_spf")
    op.drop_column("domains", "greylist")



"""CHAPPS Data Validation Models
-----------------------------

This module consists of the Pydantic_ data models required to represent and
also communicate about the various database records which control CHAPPS.
There are a few types of models, and perhaps more types to come, so some short
discussion of what they are and why they are needed seems reasonable.

Representational Models
~~~~~~~~~~~~~~~~~~~~~~~

There are (currently quite simple) objects which control CHAPPS: **User**
objects are at the root of that tree, and they are linked to one (outbound)
**Quota** object each to control how much email they may send.  **User**
objects may also be linked to multiple **Domain** and **Email** objects, to
represent who they are allowed to appear to be while sending email.

FastAPI_ uses Pydantic_ for data validation during API execution.  It behooves
us to define Pydantic_ models as representations of the objects, therefore.
This could be considered the front-facing identity of the object, which has a
back-facing identity represented by its database model, defined in
:mod:`~.dbmodels`.

A metaclass is defined which implements `__getattr__`, in order to allow the
validation model to masquerade as a database model to a certain extent.  That
routine expects the validation model to define a subclass called `Meta` with a
class attribute called `orm_model` which refers to the database model for the
object.  In this way, the validation model (class) is empowered to marshall a
set of instances from the database, without a lot of messy dereferencing.

API Response Models
~~~~~~~~~~~~~~~~~~~

In order to specify to the API constructors and the automatic API documentation
generators what the response for a particular API route should look like
(contain), more Pydantic_ models are defined.

All responses contain the CHAPPS version string and UNIX epoch time stamp in
them, as well as the response to the query, and possibly optional data
regarding an object's associations.  A fair number of response models are
defined, and they also fall into a few categories.

Unitary Data Model Responses
++++++++++++++++++++++++++++

When a single object of the primary type is being returned, that object is the
value of the `response` key in the object returned by the API.  If the object
has associations to objects of other types, the expectation is that a list of
those objects will be returned as the value of a key named for the association,
as if it were to be accessed via the ORM.  Those associated objects are listed
without any of their own associations included.

Data Model Listing Responses
++++++++++++++++++++++++++++

These response models are named almost exactly the same as their unitary
counterparts, but with their model names pluralized.  They will contain a list
of objects of the relevant type in their `response` attributes, without any
associations.

Custom Live Responses
+++++++++++++++++++++

Some of the response models are meant to relay information from the
:mod:`~.live` API routes, which deal with the current state of CHAPPS as
reflected in Redis.  These are each explained in their own documentation.

Basic Datatype Responses
++++++++++++++++++++++++

Some operations return a very simple value, such as an integer or string, and
so there are some response models to use in such cases.

.. todo::

  perhaps construct as a completely separate project a framework for creating
  arbitrary 'double-ended' Pydantic_ / SQLAlchemy_ data model objects, with
  arbitrary join tables and basic route factories for FastAPI_.  There are some
  similar projects but I couldn't find one which supported compound primary
  keys.

"""

from chapps._version import __version__
from chapps import dbmodels
from typing import Optional, List, Dict, Union, Tuple
from pydantic import BaseModel, constr, Field, validator
from pydantic.main import ModelMetaclass
from enum import Enum
import validators
import time

VERSTR = f"CHAPPS v{__version__}"  # avoiding dependency on config


class AssocOperation(str, Enum):
    """'add', 'subtract', or 'replace'

    use 'replace' only with unitary associations; logic in
    :class:`~.JoinAssoc` which responds
    to the `replace` operation is designed to work only with
    scalar values.

    """

    add = "add"
    subtract = "subtract"
    replace = "replace"


class SDAStatus(str, Enum):
    """sender-domain auth status: AUTH, PROH, or NONE"""

    AUTH = "AUTHORIZED"
    PROH = "PROHIBITED"
    NONE = "NOT CACHED"


class SPFResult(str, Enum):
    """SPF check results"""

    passing = "pass"
    fail = "fail"
    temperror = "temperror"
    permerror = "permerror"
    softfail = "softfail"
    none_neutral = "none_neutral"


class PolicyResponse:
    """A wrapper for policy results"""

    @staticmethod
    def policy_response(passing: bool, decision: str):
        """Wrap an action in order to return it as a PolicyResponse

        This routine is suitable for use as a decorator
        """

        def pr_decorator(func: callable):
            def policy_response(message="", *args, **kwargs):
                return PolicyResponse(
                    response=func(message, *args, **kwargs),
                    passing=passing,
                    decision=decision,
                )

            return policy_response

        return pr_decorator

    def __init__(
        self, *, response: str, passing: bool, decision: Optional[str] = None
    ):
        self.response = response
        self.passing = passing
        self.decision = decision

    def __getattribute__(self, attr):
        if attr in dir(str):

            def method(self, *args, **kwargs):
                return getattr(self.response, attr)(*args, **kwargs)

            return method.__get__(self)
        else:
            return super().__getattribute__(attr)

    def __getitem__(self, *args, **kwargs):
        return self.response.__getitem__(*args, **kwargs)

    def __eq__(self, *args, **kwargs):
        return self.response.__eq__(*args, **kwargs)

    def __bool__(self):
        return self.passing

    def __str__(self):
        return self.response

    def __repr__(self):
        return (
            f"PolicyResponse(response='{self.response}',"
            f" passing={self.passing}"
            + (f", decision='{self.decision}'" if self.decision else "")
            + ")"
        )


# a metaclass for passing calls through to the orm_model
class CHAPPSMetaModel(ModelMetaclass):
    """Metaclass for CHAPPS Pydantic models

    We inject an override for :meth:`~.__getattr__` in order to attempt to find
    missing attributes on the ORM class attached via the `Meta` subclass of
    each model class.  This allows the Pydantic data-model class to serve as a
    proxy for the ORM class, meaning that we can handle Pydantic models in the
    API code, and still call ORM methods on them, and cause corresponding ORM
    objects to be instantiated on demand, etc.

    .. document private functions
    .. automethod:: __getattr__

    """

    def __getattr__(cls, var):
        """ORM Masquerading

        If the requested attribute exists on the :const:`~.orm_model`, return
        it, or else `None`.  Note that while the variable name used assumes the
        attribute will refer to a callable, it will work on any attribute.

        """
        orm_model = cls.Meta.orm_model
        orm_method = getattr(orm_model, var, None)
        try:
            return orm_method or super().__getattr__(var)
        except AttributeError:
            raise AttributeError(
                f"No attribute {var} could be found on "
                + ".".join([cls.__module__, cls.__name__])
                + "or on "
                + ".".join([orm_model.__module__, orm_model.__name__])
            )


# there could be a metaclass which would look for dbmodels classes matching
# the names of subclasses of CHAPPSModel and automatically hook up their
# Meta.orm_model data ... if the number of tables starts to grow
class CHAPPSModel(BaseModel, metaclass=CHAPPSMetaModel):
    """Base API data model

    All models should define a class called `Meta` and define within it a
    variable called `orm_model` as a reference to the ORM model class
    (generally defined in :mod:`.dbmodels`) corresponding to the
    data model.  In this abstract superclass, the ORM model reference is to the
    parallel abstract ORM model superclass.

    Models also define a class called `Config`, which is used by
    :mod:`Pydantic`.  In it, `orm_mode` should be set to `True`.  It is
    also possible to include additional detail for the OpenAPI documentation
    parser.

    All models have these attributes/columns:

    """

    id: int
    """integer auto-incrementing primary identifier"""
    name: str
    """unique string label"""

    class Meta:
        """Used by CHAPPS"""

        orm_model = dbmodels.DB_Base
        """The ORM model class corresponding to this data model"""

    @classmethod
    def id_name(cls) -> str:
        """:returns: name of the ID column in a join table"""
        return "_".join(str(cls.Meta.orm_model.id).lower().split("."))

    @classmethod
    def wrap(cls, orm_instance):
        """Wrap an ORM instance in its corresponding Pydantic data model

        :param cls.Meta.orm_model orm_instance: an ORM instance of the
          appropriate type

        :returns: a pydantic model created from an ORM model

        """
        if not orm_instance:  # could be None or []
            return orm_instance
        try:
            orm_iter = iter(orm_instance)
            return [cls.from_orm(oi) for oi in orm_iter]
        except TypeError:
            return cls.from_orm(orm_instance)

    @classmethod
    def join_assoc(cls, **kwargs) -> dbmodels.JoinAssoc:
        """Create a :class:`~.JoinAssoc` with this class as the source

        :param str assoc_name: attribute name of the association

        :param type assoc_type: usually :obj:`int` or :obj:`List[int]`; should
          be the type for the API to expect when setting up the route metadata

        :param DB_Base assoc_model: a reference to the dbmodel class of the
          associated object

        :param str assoc_id: label of associated object's ID column in the
          join table

        :param sqlalchemy.schema.Table table: a reference to the join table
          schema; it will be a constant in this module, generally

        This convenience routine for generating a
        :class:`~.JoinAssoc` provides the source model and
        ID-column info as it passes on the other arguments.

        """
        # if it became necessary to track some other arbitrary id-column name
        # we could accomplish that with a metaclass, and then just set it
        # in each subclass
        return dbmodels.JoinAssoc(cls, cls.__name__.lower() + "_id", **kwargs)


class User(CHAPPSModel):
    """**User** objects represent entities authorized to send email"""

    # The **User** is central to CHAPPS's policy-enforcement strategy.
    # When a **User** attempts to send email, CHAPPS is able to check:

    # 1. That **User**\ 's **Quota**

    # 2. Whether the **User** is authorized to send email from the proposed
    #    email's apparent sender-\ **Domain**, or whether they might be
    #    authorized to send email appearing to come from the entire **Email**
    #    address.

    name: constr(strip_whitespace=True, min_length=5, max_length=127) = Field(
        ...,
        title="user identifier",
        description="may be an email address, or some other unique identifier",
    )
    """user identifiers may be from 5 to 127 chars long"""

    class Config:
        orm_mode = True
        schema_extra = dict(
            example=dict(id=0, name=("[user.identifier@]domain.name"))
        )

    class Meta:
        orm_model = dbmodels.User


class Quota(CHAPPSModel):
    """**Quota** objects represent transmission count limits"""

    # The time-interval over which **Quota** objects are enforced is 24 hr.
    # They therefore have an integer `quota` field which contains the limit
    # of transmissions per 24 hours.  A sliding window is applied to a
    # transmission-attempt history, in order to avoid having a daily reset.

    # **Quota** objects also have `id` and `name` fields, like all models.

    name: Optional[
        constr(strip_whitespace=True, min_length=3, max_length=31)
    ] = Field(
        title="quota label", description="a descriptive tag", default=None
    )
    """quota labels may be up to 31 chars long"""
    quota: Optional[int]
    """unique integer outbound transmission limit"""

    class Config:
        orm_mode = True
        schema_extra = dict(
            example=dict(id=0, name="fiftyPerHour", quota=1200)
        )

    class Meta:
        orm_model = dbmodels.Quota


class Domain(CHAPPSModel):
    """Domain objects have a name and ID; the name never contains an `@`"""

    name: Optional[
        constr(
            strip_whitespace=True, to_lower=True, min_length=5, max_length=63
        )
    ] = Field(default=None, title="domain name")
    """domain names may be up to 63 chars long"""
    greylist: Optional[bool] = Field(title="perform greylisting", default=None)
    """flag indicating whether to greylist all domain's inbound email"""
    check_spf: Optional[bool] = Field(title="enforce SPF", default=None)
    """flag indicating whether to check SPF for domain's inbound email"""

    class Config:
        orm_mode = True
        schema_extra = dict(
            example=dict(
                id=0, name="[sub.]domain.tld", greylist=False, check_spf=True
            )
        )

    class Meta:
        orm_model = dbmodels.Domain

    @validator("name")
    def domain_validator(cls, val):
        result = validators.domain(val)
        if result:
            return val
        raise ValueError("The name field must be a valid domain name.")


class Email(CHAPPSModel):
    """Email objects have a name and ID; the name always contains an `@`"""

    name: constr(strip_whitespace=True, min_length=6, max_length=127) = Field(
        ..., title="email address"
    )

    class Config:
        orm_mode = True
        schema_extra = dict(example=dict(id=0, name="someone@example.com"))

    class Meta:
        orm_model = dbmodels.Email

    @validator("name")
    def email_validator(cls, val):
        if validators.email(val):
            return val
        raise ValueError("The name field must be an email address.")


class CHAPPSResponse(BaseModel):
    """Base :mod:`Pydantic` model for API responses"""

    version: str = VERSTR
    """The CHAPPS version as a string"""
    timestamp: float = time.time()
    """When this response was generated"""
    response: object
    """Whatever piece of data was requested"""

    class Config:
        schema_extra = dict(version=VERSTR, timestamp=time.time())

    @classmethod
    def send(model, response=None, **kwargs):
        """Utility function for encapsulating responses in a standard body"""
        mkwargs = dict(version=VERSTR, timestamp=time.time())
        if response:
            mkwargs["response"] = response
        return model(**mkwargs, **kwargs)


class UserResp(CHAPPSResponse):
    """Data model for responding with a single **User** record"""

    response: User
    """The `response` field contains a **User** record"""
    domains: Optional[List[Domain]] = None
    """A list of associated **Domain** records may be included"""
    emails: Optional[List[Email]] = None
    """A list of associated **Email** records may be included"""
    quota: Optional[Quota] = None
    """The **Quota** record associated with the **User** may be included"""


class UsersResp(CHAPPSResponse):
    """Data model for responding with a list of **User** records"""

    response: List[User]
    """A list of **User** objects"""


class DomainResp(CHAPPSResponse):
    """Data model for responding with a single **Domain** record"""

    response: Domain
    """A **Domain** object"""
    users: Optional[List[User]] = None
    """A list of **User** objects associated to the **Domain** may be included
    """


class DomainsResp(CHAPPSResponse):
    """Data model for responding with a list of **Domain** records"""

    response: List[Domain]
    """A list of **Domain** objects"""


class EmailResp(CHAPPSResponse):
    """Data model for responding with a single **Email** record"""

    response: Email
    """An **Email** object"""
    users: Optional[List[User]] = None
    """A list of associated **User** objects may be included"""


class EmailsResp(CHAPPSResponse):
    """Data model for responding with a list of **Email** records"""

    response: List[Email]
    """A list of **Email** objects"""


class QuotaResp(CHAPPSResponse):
    """Data model for responding with a single **Quota** record"""

    response: Quota
    """A **Quota** object"""


class QuotasResp(CHAPPSResponse):
    """Data model for responding with a list of **Quota** records"""

    response: List[Quota]
    """A list of **Quota** objects"""


class IntResp(CHAPPSResponse):
    """Data model for responding with an integer"""

    response: int
    """An integer"""


class FloatResp(CHAPPSResponse):
    """Data model for responding with a float"""

    response: float
    """A floating-point number"""


class TextResp(CHAPPSResponse):
    """Data model for responding with a string"""

    response: str
    """A string"""


class TimeResp(FloatResp):
    """Data model for responding with a UNIX epoch time value"""

    response: float = time.time()
    """UNIX epoch time (UTC)"""


class InstanceTimesResp(CHAPPSResponse):
    """Data model for returning a list of instances and timestamps"""

    response: List[Tuple[str, float]]
    """A list of (instance, timestamp) tuples"""

    class Config:
        schema_extra = {
            "example": {
                "response": [
                    ["instance001", time.time() - 12000.0],
                    ["instance002", time.time() - 6000.0],
                ],
                "timestamp": time.time(),
                "version": VERSTR,
            }
        }


class LiveQuotaResp(CHAPPSResponse):
    """Data model for responses from the Live API"""

    response: int
    """An integer"""
    remarks: List[str] = []
    """A list of string remarks may be included"""


class SourceUserMapResp(CHAPPSResponse):
    """
    A source-user map is a dict-of-dicts:
      - top level key is domain or email name
      - second key is user
      - value is SDAStatus
    """

    response: Dict[str, Dict[str, SDAStatus]]
    """A map of auth-subject to dicts of username mapping to status"""


class BulkQuotaResp(CHAPPSResponse):
    """
    Maps **User** `name` onto **Quota** `id`

    With descriptive labels and optional remarks (from the live API)
    """

    response: List[Dict[str, Union[str, int, None]]]
    remarks: List[str] = []


class BulkDomainsResp(CHAPPSResponse):
    """
    Maps **User** `name` onto lists of **Domain** `id`

    With descriptive labels and optional remarks (from the live API)
    """

    response: List[Dict[str, Union[str, List[int], None]]]
    remarks: List[str] = []


class BulkEmailsResp(BulkDomainsResp):
    """
    Maps **User** `name` onto lists of **Email** `id`

    With descriptive labels and optional remarks (from the live API)
    """


class DeleteResp(TextResp):
    """Data model for responding to deletion requests"""

    response: str = "deleted"
    """The string 'deleted'"""


# the following classes are somewhat speculative for now
# class FloatResp(CHAPPSResponse):
#     """Data model for responding with a float"""

#     response: float
#     """A float"""

user_quota_assoc = User.join_assoc(
    assoc_name="quota",
    assoc_type=int,
    assoc_model=Quota,
    assoc_id=Quota.id_name(),
    table=User.Meta.orm_model.metadata.tables["quota_user"],
)

user_domains_assoc = User.join_assoc(
    assoc_name="domains",
    assoc_type=List[int],
    assoc_model=Domain,
    assoc_id=Domain.id_name(),
    table=User.Meta.orm_model.metadata.tables["domain_user"],
)

user_emails_assoc = User.join_assoc(
    assoc_name="emails",
    assoc_type=List[int],
    assoc_model=Email,
    assoc_id=Email.id_name(),
    table=User.Meta.orm_model.metadata.tables["email_user"],
)

domain_users_assoc = Domain.join_assoc(
    assoc_name="users",
    assoc_type=List[int],
    assoc_model=User,
    assoc_id="user_id",
    table=Domain.Meta.orm_model.metadata.tables["domain_user"],
)

email_users_assoc = Email.join_assoc(
    assoc_name="users",
    assoc_type=List[int],
    assoc_model=User,
    assoc_id="user_id",
    table=Email.Meta.orm_model.metadata.tables["email_user"],
)

# there is no quota_users_assoc because we don't list users by quota association



"""
InboundPPR
-----------

And possibly other classes related to outbound traffic, but right now there is
only one, which is an outbound-only subclass of
:py:class:`chapps.util.PostfixPolicyRequest`.

"""
from typing import List
from functools import cached_property
from chapps.util import PostfixPolicyRequest

# from chapps.config import config, CHAPPSConfig
from chapps.signals import NoRecipientsException
import logging

logger = logging.getLogger(__name__)


class InboundPPR(PostfixPolicyRequest):
    """Encapsulates logic to identify the client domain for inbound mail
    """

    def __init__(self, payload: List[str], **kwargs):
        """Create a new inbound policy request"""
        super().__init__(payload)

    @cached_property
    def recipient_domain(self):
        if not len(self.recipients):
            raise NoRecipientsException(
                f"PPR {ppr.instance} contains no recipients"
            )
        domains = set([self.domain_from(e) for e in self.recipients])
        if len(domains) > 1:
            # raise MultipleInboundRecipientsException ?
            logger.debug(
                f"Using first recipient {ppr.recipients[0]} for domain flags."
            )
        return self.domain_from(self.recipients[0])

    def __str__(self):
        """In certain contexts, `str(<o_ppr>)` is used for brevity

        The routine tries to use the `user` which is the point of the class,
        but if it cannot determine a non-nil user name, it falls back to
        printing a bit of extra detail.

        """
        try:
            return (
                f"i={self.instance} "
                f"domain={self.recipient_domain} "
                f"sender={self.sender or 'None'} "
                f"client_address={self.client_address} "
                f"recipient={self.recipient}"
            )
        except Exception:
            return (
                f"i={self.instance} "
                f"sender={self.sender or 'None'} "
                f"client_address={self.client_address} "
                f"recipient={self.recipient}"
            )



#!/usr/bin/env python3
"""CHAPPS database setup script"""
### This script reads the CHAPPS config and uses the database information contained therein
###   to initialize the database.  It should be idempotent, and safe to run against an existing
###   config database.

from chapps.dbsession import sql_engine
from chapps.dbmodels import DB_Base


def setup_chapps_database():
    DB_Base.metadata.create_all(sql_engine)


if __name__ == "__main__":
    setup_chapps_database()



"""Use Alembic to bring migrations up to date"""
# Python script that will apply the migrations up to head
# from https://stackoverflow.com/questions/42383400/python-packaging-alembic-migrations-with-setuptools
import alembic.config
import os

here = os.path.dirname(os.path.abspath(__file__))

alembic_args = ["-c", os.path.join(here, "alembic.ini"), "upgrade", "head"]


def main():
    alembic.config.main(argv=alembic_args)



"""API route implementations

Each major category of route gets its own router.

Much of the actual code is contained in the :mod:`common` module,
which contains factories for building the routes related to object
manipulation.  The :mod:`live` module contains routes for interacting
with Redis, which represents the live state of CHAPPS.

"""
__all__ = ["common", "email", "users", "quotas", "domains", "live"]



"""
OutboundPPR
-----------

And possibly other classes related to outbound traffic, but right now there is
only one, which is an outbound-only subclass of
:py:class:`chapps.util.PostfixPolicyRequest`.

"""
from typing import List
from chapps.util import PostfixPolicyRequest
from chapps.config import CHAPPSConfig
from chapps.signals import ConfigurationError, AuthenticationFailureException
import logging

logger = logging.getLogger(__name__)


class OutboundPPR(PostfixPolicyRequest):
    """Encapsulates logic to identify the authenticated user sending the email.

    Ignoring non-authenticated email attempts, which I think everyone does
    today, outbound requests should all be associated to a known user who has
    authenticated to Postfix.  Postfix provides information in the payload
    which can be used to identify the user.  That piece of data may then be
    used to apply policies to the outbound email of a particular user.

    In order to allow for the possibility that other sites might prefer to use
    arbitrary bits of the Postfix policy request payload to identify users,
    CHAPPS allows for the key used to extract that data to be specified in the
    config file.  At present, there is no support for combining more than one
    key.

    .. admonition:: Notes about subclassing

       Don't forget that the normal attribute space is reserved for the payload
       data.  All internal attributes should start with `_` (underscore).

    .. todo::

      Arbitrary user-key composition functionality could be achieved by a
      classmethod setting the user-getter routine, accepting as its argument a
      closure which accepts a :class:`~chapps.util.PostfixPolicyRequest` and
      returns a string which is the user-identifier.

    """

    _memoized_routines = dict()

    def __init__(self, payload: List[str], *, cfg: CHAPPSConfig = None):
        """Create a new outbound policy request"""
        super().__init__(payload)
        self._config = cfg or CHAPPSConfig.get_config()
        self._params = self._config.chapps
        if cfg:
            logger.debug(
                "Got override config from file: " + cfg.chapps.config_file
            )
        else:
            logger.debug(
                "Using global config based on: " + self._params.config_file
            )

    @classmethod
    def clear_memoized_routines(cls) -> None:
        """Clear all memoized routines

        This mainly exists to facilitate testing.

        """
        cls._memoized_routines.clear()

    def __str__(self):
        """In certain contexts, `str(<o_ppr>)` is used for brevity

        The routine tries to use the `user` which is the point of the class,
        but if it cannot determine a non-nil user name, it falls back to
        printing a bit of extra detail.

        """
        try:
            return (
                f"i={self.instance} "
                f"user={self.user} "
                f"sender={self.sender or 'None'} "
                f"client_address={self.client_address} "
                f"recipient={self.recipient}"
            )
        except Exception:
            return (
                f"i={self.instance} "
                f"sasl_username={self.sasl_username or 'None'} "
                f"ccert_subject={self.ccert_subject or 'None'} "
                f"sender={self.sender or 'None'} "
                f"client_address={self.client_address} "
                f"recipient={self.recipient}"
            )

    @property
    def user(self) -> str:
        """Return and memoize the user-identifier.

        :returns: the user-identifier

        :rtype: str

        :raise AuthenticationFailureException: when no user-identifier can be
          found, and the `require_user_key` setting of the `[CHAPPS]` section
          of the config is set to :obj:`True`

        :raise ValueError: when no user-identifier is found, but user keys
          are not required

        The underlying routine raises :exc:`ValueError` when no user-identifier
        can be found.  If the config stipulates that a user-identifier must be
        found, this routine raises :exc:`AuthenticationFailureException` to
        signal that the email originated from a source which did not
        authenticate.

        """
        if not "_user" in self.__dict__:
            try:
                self._user = self._get_user()
            except ValueError as e:
                if self._params.require_user_key:
                    raise AuthenticationFailureException()
                else:
                    raise e
        return self._user

    # this routine creates and memoizes a routine if need be
    # it also uses that routine to return the value to be used to track a user
    def _get_user(self) -> str:
        """Obtain the user value, and memoize the procedure

        In an attempt to support as many different configuration scenarios as
        possible, the codebase can attempt to find a user-identifier in a few
        different places, in a search path.  Since it is possible to configure
        this search path via the config file, the actual search function is
        produced as a closure and memoized at the class level, so that all
        future instances will be able to use the same function without creating
        a new (identical) closure first.

        The closure produced by this factory raises :exc:`ValueError` if it
        cannot find a non-nil user-identifier.

        .. note::

           Any alternate closure provided for this purpose should also raise
           :exc:`ValueError` if no user-identifier can be found, as it is
           handled explicitly in :meth:`.user`

        :meta public:

        """
        # see if we already have a procedure from some previous iteration
        get_user = self.__class__._memoized_routines.get("get_user", None)
        cfg = self._params
        # if there is no procedure, we build one
        # logger.debug("Using config file: " + cfg.config_file)
        if not get_user:
            if cfg.require_user_key:
                if not cfg.user_key:
                    raise ConfigurationError(
                        (
                            "If require_user_key is True, "
                            "then user_key must be set."
                        )
                    )
                qk_list = [cfg.user_key]
                # logger.debug(
                #     f"User key required; {cfg.user_key} must appear in PPRs."
                # )
            else:
                qk_list = [
                    "sasl_username",
                    "ccert_subject",
                    "sender",
                    "client_address",
                ]
                qk = cfg.user_key
                if qk and qk != qk_list[0]:
                    qk_list = [qk, *qk_list]
                # logger.debug(
                #     "User key not required.  Using search path: "
                #     + (":".join(qk_list))
                # )

            def get_user(ppr):
                for k in qk_list:
                    user = getattr(ppr, k, None)
                    if user and user != "None":
                        # logger.debug(
                        #     f"Selecting user-identifier {user} from key {k}"
                        # )
                        return user
                raise ValueError(
                    (
                        "None of the following keys had values in "
                        f"the provided PPR: {qk_list}"
                    )
                )

            # this memoizes the procedure for finding the user
            self.__class__._memoized_routines["get_user"] = get_user
        # execute the procedure and return the result
        return get_user(self)



"""
**Email** record management implemented by factories
----------------------------------------------------

This module defines the management routes for **Email** records, and also the :class:`~.JoinAssoc` between **Email** and **User** tables.

It is implemented using factory functions from the :mod:`~.common` module, in a nearly-identical way to the :mod:`~.domains` module, just using :class:`~chapps.models.Email` as the main data model.

"""

from typing import List
from starlette import status
from fastapi import APIRouter  # , Body, Path, HTTPException
from chapps.models import (
    User,
    Email,
    EmailResp,
    EmailsResp,
    UsersResp,
    DeleteResp,
    TextResp,
    AssocOperation,
    email_users_assoc,
)
from chapps.rest.routers.common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
    adjust_associations,
    list_associated,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

api = APIRouter(
    prefix="/emails",
    tags=["emails"],
    responses={status.HTTP_404_NOT_FOUND: {"description": "Email not found."}},
)
"""The **Email** object API router"""

email_join_assoc = [email_users_assoc]
"""List of join associations for **Email** objects"""

api.get("/", response_model=EmailsResp)(
    list_items(Email, response_model=EmailsResp)
)

api.get("/{item_id}", response_model=EmailResp)(
    get_item_by_id(Email, response_model=EmailResp, assoc=email_join_assoc)
)

api.get("/{item_id}/users/", response_model=UsersResp)(
    list_associated(Email, assoc=email_users_assoc, response_model=UsersResp)
)

api.post(
    "/",
    status_code=201,
    response_model=EmailResp,
    responses={status.HTTP_409_CONFLICT: {"description": "Unique key error."}},
)(
    create_item(
        Email,
        response_model=EmailResp,
        params=dict(name=str),
        assoc=email_join_assoc,
    )
)

api.put("/", response_model=EmailResp)(
    update_item(Email, response_model=EmailResp, assoc=email_join_assoc)
)

api.put("/{item_id}/users/", response_model=TextResp)(
    adjust_associations(
        Email, assoc=email_join_assoc, assoc_op=AssocOperation.add
    )
)

api.delete("/{item_id}/users/", response_model=TextResp)(
    adjust_associations(
        Email, assoc=email_join_assoc, assoc_op=AssocOperation.subtract
    )
)

api.delete(
    "/",
    response_model=DeleteResp,
    responses={
        status.HTTP_202_ACCEPTED: {"description": "Items will be deleted."},
        status.HTTP_204_NO_CONTENT: {"description": "No item to delete."},
        status.HTTP_409_CONFLICT: {
            "description": "Database integrity conflict."
        },
    },
)(
    delete_item(Email)
)  # params=dict(ids=List[int]) by default



"""
SPF Enforcement policy manager
------------------------------

Isolated here to prevent the core codebase from depending upon the SPF
libraries.

"""
import spf
from chapps.signals import NoRecipientsException, NoSuchDomainException
from chapps.policy import InboundPolicy, PostfixActions, GreylistingPolicy
from chapps.util import PostfixPolicyRequest
from chapps.inbound import InboundPPR
from chapps.config import CHAPPSConfig
import logging

logger = logging.getLogger(__name__)


class PostfixSPFActions(PostfixActions):
    """
    Postfix Action translator for :py:class:`chapps.policy.SPFEnforcementPolicy`

    .. caution::

        The SPF functionality of CHAPPS is not considered complete.  YMMV

    """

    def greylist_factory(self):
        greylisting_policy = GreylistingPolicy(self.config)
        greylisting_actions = greylisting_policy.actions
        spf_policy = self.spf_policy

        def greylist(msg, *args, **kwargs):
            """Greylisting closure

            This method is meant to share the same signature as the other
            action methods, mainly defined on
            :py:class:`chapps.policy.PostfixActions`

            The `greylist` action causes the email in question to be
            greylisted, according to the policy.  The `msg` is used as the
            message for the Postfix directive, or if the message has no
            contents and the email is being deferred, the string "due to SPF
            enforcement policy" is used.  Because the greylisting policy's
            message is prepended later, the actual message delivered by Postfix
            will look something like "greylisted due to SPF enforcement
            policy".

            """
            ppr = kwargs.get("ppr", None)
            if ppr is None:
                raise ValueError(
                    "PostfixSPFActions.greylist() expects a ppr= kwarg "
                    "providing the PPR for greylisting."
                )
            if greylisting_policy.approve_policy_request(ppr, force=True):
                passing = spf_policy.actions.action_for("pass")
                return passing(msg, ppr, *args, **kwargs)
            if len(msg) == 0:
                msg = "due to SPF enforcement policy"
            return greylisting_actions.fail(msg, ppr, *args, **kwargs)

        return greylist

    _cached_actions = {}
    _action_factories = dict(greylist=greylist_factory)

    def __getattr__(self, attr):
        if attr not in self._cached_actions:
            if attr in self._action_factories:
                self._cached_actions[attr] = self._action_factories[attr](self)
            else:
                return super().__getattr__(attr)
        return self._cached_actions[attr]

    def __init__(self, spf_policy, cfg: CHAPPSConfig = None):
        """Create a new PostfixSPFActions instance

        :param ~.SPFEnforcementPolicy spf_policy: The SPF policy object which
          instantiated this actions object

        :param cfg: Optional config override

        The config is normally taken from the `~.SPFEnforcementPolicy` passed
        in as the first argument, but that config may be overridden by
        providing the config argument.

        """
        super().__init__(cfg or spf_policy.config)
        self.spf_policy = spf_policy
        self.params = self.config.actions_spf

    def _mangle_action(self, action):
        """
        Perform additional mangling to turn either of 'none' or 'neutral' into
        the same symbol, 'none_neutral'

        """
        if action == "none" or action == "neutral":
            action = "none_neutral"
        else:
            action = super()._mangle_action(action)
        return action

    def action_for(self, spf_result):
        """

        Override
        :py:meth:`chapps.policy.PostfixActions.action_for()` to provide
        action closures for the different SPF results.  The closures are
        memoized, so that they only need be constructed once per runtime.

        """
        spf_result = self._mangle_action(spf_result)
        action = getattr(self, spf_result, None)
        if action:
            return action
        return self._get_closure_for(spf_result)  #  this memoizes its result


class SPFEnforcementPolicy(InboundPolicy):
    """Policy manager which enforces SPF policy

    Instance attributes (in addition to those
    of :class:`~chapps.policy.EmailPolicy`):

      :actions: a :class:`~.PostfixSPFActions` instance

    Behavior of the SPF enforcer is configured under the
    `[PostfixSPFActions]` heading in the config file.

    """

    redis_key_prefix = "spf"
    """For option flag in Redis"""

    def __init__(self, cfg=None):
        """Setup an SPF enforcement policy manager

        :param chapps.config.CHAPPSConfig cfg: optional config override

        """
        super().__init__(cfg)
        self.actions = PostfixSPFActions(self)

    def acquire_policy_for(self, ppr: InboundPPR) -> bool:
        """Determine whether this PPR's recipient wants SPF enforcement

        :param ppr: A PPR representing the email

        :raises: :class:`~chapps.signals.NoSuchDomainException` if the
          recipient domain is not present in the configuration RDBMS

        """
        with self._adapter_handle() as adapter:
            result = adapter.check_spf_on(ppr.recipient_domain)
        self._store_control_data(ppr.recipient_domain, result)
        return result

    def _get_control_data(self, ppr: InboundPPR) -> int:
        option_key = self.domain_option_key(ppr)
        option_bits = self.redis.get(option_key)
        if option_bits is None:
            return None
        else:
            return int(option_bits)

    def enabled(self, ppr: InboundPPR) -> bool:
        """Main entry point to determining whether to enforce SPF

        :param ppr: A PPR representing the email

        Returns a boolean indicating whether to enforce SPF for this
        recipient.
        """
        option_flag = None
        try:
            option_flag = self._get_control_data(ppr)
        except NoRecipientsException:
            logger.exception(f"No recipient in PPR {ppr.instance}")
            return False
        except Exception:
            logger.exception("UNEXPECTED")
            return False
        if option_flag is None:
            try:
                option_flag = self.acquire_policy_for(ppr)
            except NoSuchDomainException:
                logger.exception(f"No domain matched {ppr.recipient}")
                return False
        option_flag = int(option_flag)
        return option_flag == 1

    def _approve_policy_request(self, ppr: PostfixPolicyRequest) -> str:
        """Perform SPF enforcement decision-making

        :param ppr: a Postfix payload

        :returns: a string which contains a Postfix instruction

        The :class:`~.PostfixSPFActions` class translates
        between the outcome of the SPF check and the configured response
        thus indicated, which gets sent back to Postfix.

        .. todo::

          Allow configuration of the list of results of the HELO check which
          should be honored, rather than resulting in proceeding to the MAIL
          FROM check.  Currently, only `fail` is honored, meaning that any
          other result will mean that a MAIL FROM check is to be conducted, and
          its result used as the result.

        """
        # Bypass if whitelisted -- we don't even want a header
        if self._whitelisted(ppr):
            return "DUNNO"
        # First, check the HELO name
        helo_sender = "postmaster@" + ppr.helo_name
        query = spf.query(
            ppr.client_address,
            helo_sender,
            ppr.helo_name,
            querytime=self.params.spf_query_timeout,
        )
        result, _, message = query.check()
        if result in [
            "fail"
        ]:  # TODO: allow configuration of HELO results to honor
            action = self.actions.action_for(result)
        else:
            # the HELO name did not produce a definitive result, so check MAILFROM
            query = spf.query(ppr.client_address, ppr.sender, ppr.helo_name)
            result, _, message = query.check()
            action = self.actions.action_for(result)
        if not self.enabled(ppr):  # prepend headers even w/o enforcement
            action = self.actions.prepend
        return action(
            message,
            ppr=ppr,
            prepend="Received-SPF: " + query.get_header(result, "spfquery"),
        )



"""
**Domain** record management implemented by factories
-----------------------------------------------------

This module defines the various routes for manipulating **Domain** records.  It
also defines the :class:`~.dbmodels.JoinAssoc` between **Domain** and **User**
tables.

Normally, each API route would be a decorated function or coroutine definition.
However, in this case, all the routes are created by factories from
:mod:`.common`.  This makes it impossible to attach docstrings.

In order to provide basic documentation and some examples how to use those
:mod:`~.common` API route coroutine factories, the routes will be described
below, as part of the `.api` docstring.

"""
from typing import List, Optional
from starlette import status
from fastapi import APIRouter  # , Body, Path, HTTPException
from chapps.models import (
    User,
    Domain,
    DomainResp,
    DomainsResp,
    UsersResp,
    DeleteResp,
    TextResp,
    AssocOperation,
    domain_users_assoc,
)
from chapps.rest.routers.common import (
    get_item_by_id,
    list_items,
    create_item,
    delete_item,
    update_item,
    adjust_associations,
    list_associated,
)
import logging
import chapps.logging

logger = logging.getLogger(__name__)
logger.setLevel(chapps.logging.DEFAULT_LEVEL)

api = APIRouter(
    prefix="/domains",
    tags=["domains"],
    responses={
        status.HTTP_404_NOT_FOUND: {"description": "Domain not found."}
    },
)
"""`FastAPI`_ :class:`~fastapi.APIRouter` object for domains

I will include a few examples here, rather than copy in all of the code.  It
seems important to provide and annotate at least a few examples in order to
ensure that others, or my later self, will understand how this is all intended
to work.

First, a simple example: the listing route.  **Domain** objects, like just
about any object represented this way, need to be listable by API so that they
can be viewed by a user before being modified, linked to other objects,
queried, destroyed, etc.  In :mod:`chapps.rest.routers.common`, coroutine
factories are defined to create all of these API routes.

.. _listing-domains:
.. rubric:: Listing Domains

Normally this would look like the route definitions in the
:mod:`~.live` module, a decorated coroutine definition.  Something like this:

.. code:: python

  @api.get("/", response_model=DomainsResp)
  async def list_domains(
    item_id: int, qparams: dict = Depends(list_query_params)
  ):
      ... some code ...
      return DomainsResp.send(response, [...])

Instead, we invoke the decorator directly as a function.  As a
decorator-with-arguments, its return value is a closure which is the actual
decorator.  That real decorator expects our route function or coroutine as an
argument.  A factory from :mod:`~.common` is used to create that route
coroutine.

Therefore, the actual route is created on the `APIRouter` like so:

.. code:: python

  api.get("/", response_model=DomainsResp)(
       list_items(Domain, response_model=DomainsResp)
  )

Please note that the main data model class is provided.  All of the routes in
this module regard **Domain** as the main data model.  Because this route is
for listing objects/records, it uses a response type which includes a list.

What about a route which makes changes, especially one which is sensitive to the model's associations to other models?  Things become a little bit more complicated, but not by much.

.. _updating-domains:
.. rubric:: Updating Domains

To update a **Domain**, we need to receive the data within a **Domain** object,
plus optionally a list of **User** IDs, to associate with the **Domain**
instead of whatever is currently associated with the **Domain**.

.. code:: python

  api.put("/", response_model=DomainResp)(
      update_item(Domain, response_model=DomainResp, assoc=domain_join_assoc)
  )

In order to specify the associations to support in the route, a list of
:class:`~.JoinAssoc` objects may be provided to the `assoc` parameter.

.. todo::

  There is a further opportunity for automation through metaprogramming, to set
  up all the expected, basic routes for a given data model in a single function
  call.

(The following parameters and return type correspond to the :class:`~fastapi.APIRouter` instance, which is callable.  Sphinx autodoc makes it quite difficult to suppress this output.)

"""

domain_join_assoc = [domain_users_assoc]
"""Join association list for **Domain** objects"""

api.get("/", response_model=DomainsResp)(
    list_items(Domain, response_model=DomainsResp)
)

api.get("/{item_id}", response_model=DomainResp)(
    get_item_by_id(Domain, response_model=DomainResp, assoc=domain_join_assoc)
)

api.get("/{item_id}/users/", response_model=UsersResp)(
    list_associated(Domain, assoc=domain_users_assoc, response_model=UsersResp)
)

api.post(
    "/",
    status_code=201,
    response_model=DomainResp,
    responses={status.HTTP_409_CONFLICT: {"description": "Unique key error."}},
)(
    create_item(
        Domain,
        response_model=DomainResp,
        params=dict(
            name=str, greylist=Optional[bool], check_spf=Optional[bool]
        ),
        defaults=dict(greylist=False, check_spf=False),
        assoc=domain_join_assoc,
    )
)

api.put("/", response_model=DomainResp)(
    update_item(Domain, response_model=DomainResp, assoc=domain_join_assoc)
)

api.put("/{item_id}/users/", response_model=TextResp)(
    adjust_associations(
        Domain, assoc=domain_join_assoc, assoc_op=AssocOperation.add
    )
)

api.delete("/{item_id}/users/", response_model=TextResp)(
    adjust_associations(
        Domain, assoc=domain_join_assoc, assoc_op=AssocOperation.subtract
    )
)

api.delete(
    "/",
    response_model=DeleteResp,
    responses={
        status.HTTP_202_ACCEPTED: {"description": "Items will be deleted."},
        status.HTTP_409_CONFLICT: {
            "description": "Database integrity conflict."
        },
    },
)(delete_item(Domain))
